#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <algorithm>
#include "plugin.hpp"

#include "types.hpp"
#include "types_isd.hpp"
#include "tools.hpp"

extern PluginStartupInfo startupInfo;

wchar_t* cropLongName(wchar_t* newName, const wchar_t* name, int noChars)
{
  int size = lstrlenW(name);
  if(size > noChars)
  {
    if(name[1] == L':' && name[2] == L'\\')
    {
      lstrcpynW(newName, name, 4);
      lstrcatW(newName, L"...");
      lstrcatW(newName, name+size-noChars+6);
    }
    else
    {
      lstrcpyW(newName, L"...");
      lstrcatW(newName, name+size-noChars+3);
    }
  }
  else
    lstrcpyW(newName, name);
  return newName;
}

static char invalidChars[] = { ' ', '.', '\\', '/', ':', '?', '*', '<', '>', '|' };

bool isValidCharIsdos(char ch)
{
  if((unsigned char)ch >= 0x80 && (unsigned char)ch <= 0xAF) return true;
  if((unsigned char)ch >= 0xE0 && (unsigned char)ch <= 0xF1) return true;
  if((unsigned char)ch < 0x20 || (unsigned char)ch > 0x7F) return false;
  if(std::any_of(std::cbegin(invalidChars), std::cend(invalidChars), [ch](char x) { return x == ch; }))
      return false;
  return true;
}

bool isValidCharIsdos(wchar_t ch)
{
    if(std::any_of(std::cbegin(invalidChars), std::cend(invalidChars), [ch](wchar_t x) { return x == ch; }))
        return false;
    return true;
}

wchar_t* makeFullName(wchar_t* fullName, const wchar_t* path, const wchar_t* name)
{
  lstrcpyW(fullName, path);
  addEndSlash(fullName);
  return lstrcatW(fullName, name);
}

void makeCompatibleFileName(wchar_t* newName, const wchar_t* oldName)
{
  if(*oldName == 0) return;
  wchar_t* p = newName;
  const wchar_t* q = oldName;
  for(int i = 0; i < 8; ++i)
  {
    if(*q == 0 || *q == '.') break;

    *p = isValidCharIsdos(*q) ? *q : '_';

    ++p;
    ++q;
  }
  
  q = pointToExt(oldName);
  
  if(q)
  {
    *p++ = '.';
    for(int i = 0; i < 3; ++i)
    {
      if(*q == 0) break;

      *p = isValidCharIsdos(*q) ? *q : '_';

      ++p;
      ++q;
    }
  }
  *p = 0;
}

void makeCompatibleFolderName(wchar_t* newName, const wchar_t* oldName)
{
  if(*oldName == 0) return;
  wchar_t* p = newName;
  const wchar_t* q = oldName;
  for(int i = 0; i < 8; ++i)
  {
    if(*q == 0 || *q == '.') break;

    *p = isValidCharIsdos(*q) ? *q : '_';

    ++p;
    ++q;
  }
  *p = 0;
}
