#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <iterator>
#include <memory>
#include <string>

#include "plugin.hpp"
#include "types.hpp"
#include "detector_xlook.hpp"
#include "association.hpp"
#include "tools.hpp"
#include "tools_xlook.hpp"

#include "debug.hpp"

#define _export

void addEndSlash(std::wstring &path);

extern Association _associations[256];
extern int         _no_assosiations;

PluginStartupInfo startupInfo;
static FarStandardFunctions FSF;

xlook::DetectorManager* manager ;//= 0;

// global options should be in the registry
static const wchar_t* open_in_viewer_prefix = L"xv";
static const wchar_t* open_in_editor_prefix = L"xe";
//char* export_to_file_prefix = "xx";

static DWORD strip_hobeta_header = 1;

// folder for all temporary files
static wchar_t tmp_folder[300];

struct zxHandle
{
    wchar_t fileName[300];// 
    BYTE *data;// 
    DWORD dataSize;// 
    wchar_t ext[5];//,   detector
    char name[9];//  HoBeta 
};

extern const GUID MainGuid = __uuidof(struct __declspec(uuid("{38F55534-2393-4BE7-B269-450F2CD48A7C}")) TMainGuid);

static bool extract(xlook::Detector *dtr, const BYTE *source, DWORD sourceSize, HANDLE dest, DWORD internal_id)
{
    DWORD noUnpackedBytes = 65536;
    BYTE *unpackedBuffer = (BYTE*)malloc(noUnpackedBytes);

    bool result = dtr->extract((HoHdr*)source, source+sizeof(HoHdr), sourceSize-sizeof(HoHdr), unpackedBuffer, &noUnpackedBytes, internal_id);
    if(!result && noUnpackedBytes > 65536)
    {
        unpackedBuffer = (BYTE*)realloc(unpackedBuffer, noUnpackedBytes);
        result = dtr->extract((HoHdr*)source, source+sizeof(HoHdr), sourceSize-sizeof(HoHdr), unpackedBuffer, &noUnpackedBytes, internal_id);
    }

    if(result)
    {
        DWORD noBytesWritten;
        WriteFile(dest, unpackedBuffer, noUnpackedBytes, &noBytesWritten, 0);
    }
    
    free(unpackedBuffer);
    return result;
}

void WINAPI _export SetStartupInfoW(const struct PluginStartupInfo *info)
{
    startupInfo = *info;
    FSF = *(info->FSF);
    startupInfo.FSF = &FSF;

    // load Detectors
    manager = new xlook::DetectorManager;

    wchar_t tmp_path[300];
    GetTempPathW(std::size(tmp_path), tmp_path);
    GetTempFileNameW(tmp_path, L"!xf", 0, tmp_folder);
    DeleteFileW(tmp_folder);

    read_associations(startupInfo.ModuleName);

    lstrcatW(tmp_folder, L"\\");
    CreateDirectoryW(tmp_folder, 0);
}

static void CleanTempFolder(const wchar_t *name)
{
    wchar_t all_files[300];
    lstrcpyW(all_files, name);
    lstrcatW(all_files, L"*.*");

    WIN32_FIND_DATAW data;
    HANDLE foundFile = FindFirstFileW(all_files, &data);

    int done = (foundFile == INVALID_HANDLE_VALUE);

    while(!done)
    {
        if(lstrcmpW(data.cFileName, L".") && lstrcmpW(data.cFileName, L".."))
        {
            wchar_t file_name[300];
            lstrcpyW(file_name, name);
            lstrcatW(file_name, data.cFileName);
            if(data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
            {
                lstrcatW(file_name, L"\\");
                CleanTempFolder(file_name);
            }
            else
                DeleteFileW(file_name);
        }
        done = !FindNextFileW(foundFile, &data);
    }
    FindClose(foundFile);

    RemoveDirectoryW(name);
}

void WINAPI _export ExitFARW()
{
    CleanTempFolder(tmp_folder);

    delete manager;
}

void WINAPI _export GetPluginInfoW(struct PluginInfo *info)
{
    if(!info) return;

    info->StructSize = sizeof(*info);

    info->Flags = PF_FULLCMDLINE;

    static wchar_t prefixes[100];
    lstrcpyW(prefixes, open_in_viewer_prefix);
    lstrcatW(prefixes, L":");
    lstrcatW(prefixes, open_in_editor_prefix);
    /*lstrcat(prefixes, ":");
    lstrcat(prefixes, export_to_file_prefix);*/

    info->CommandPrefix = prefixes;
}

void WINAPI _export GetOpenPanelInfoW(struct OpenPanelInfo *Info)
{
    Info->StructSize = sizeof(*Info);
    Info->Flags      = OPIF_ADDDOTS;
    Info->HostFile   = ((zxHandle*)Info->hPanel)->fileName;
    //Info->StartPanelMode = '2';

    static wchar_t panelTitle[MAX_PATH];
    lstrcpyW(panelTitle, L" xLook:");
    lstrcatW(panelTitle, pointToName(((zxHandle*)Info->hPanel)->fileName));
    lstrcatW(panelTitle, L" ");
    Info->PanelTitle = panelTitle;
}

intptr_t  WINAPI _export GetFindDataW(struct GetFindDataInfo *Info)
{
    PluginPanelItem *item = new PluginPanelItem[1]{};

    Info->PanelItem = item;
    Info->ItemsNumber = 1;

    auto zx = (zxHandle *)Info->hPanel;

    auto PcName = new wchar_t[8 + 1 + 3 + 1]{};
    makePCName(zx->name, PcName);

    if(zx->ext[0])
    {
        lstrcatW(PcName, zx->ext);
    }
    else
    {
        lstrcatW(PcName, L".bin");
    }
    item->FileName = PcName;
    item->FileSize = zx->dataSize;

    return TRUE;
}

void WINAPI _export FreeFindDataW(const struct FreeFindDataInfo *Info)
{
    delete[] Info->PanelItem->FileName;
    delete[] Info->PanelItem;
}

intptr_t WINAPI _export GetFilesW(struct GetFilesInfo *Info)
{
    auto PanelItem{ Info->PanelItem };
    auto ItemsNumber{ Info->ItemsNumber };
    auto DestPath{ Info->DestPath };
    auto OpMode{ Info->OpMode };
    if(ItemsNumber !=1)
        return 0;

    wchar_t tmp_file_name[300];
    lstrcpyW(tmp_file_name, DestPath);
    if(*tmp_file_name && tmp_file_name[lstrlenW(tmp_file_name)-1] != L'\\') lstrcatW(tmp_file_name,L"\\");

    if(!(OpMode & OPM_SILENT))
    {
        wchar_t export_to[64] = L"&Copy ";
        lstrcatW(export_to, PanelItem[0].FileName);
        lstrcatW(export_to, L" to:");
        FarDialogItem dialogItems[6]
        {
            {DI_DOUBLEBOX, 3, 1, 52, 6, { FALSE }, nullptr,            nullptr,                                   0, L"Copy" },
            {DI_TEXT,      5, 2,  0, 0, { FALSE }, nullptr,            nullptr,                                   0, export_to },
            {DI_EDIT,      5, 3, 50, 0,  { TRUE }, L"xLook_export_to", nullptr,        DIF_HISTORY | DIF_EDITEXPAND, tmp_file_name },
            {DI_TEXT,      0, 4,  0, 0, { FALSE }, nullptr,            nullptr,                       DIF_SEPARATOR, L"" },

            {DI_BUTTON,   11, 5,  0, 0, { FALSE }, nullptr,            nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, L"&Ok" },
            {DI_BUTTON,   35, 5,  0, 0, { FALSE }, nullptr,            nullptr,                     DIF_CENTERGROUP, L"Ca&ncel" }
        };

        std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                         &__uuidof(struct __declspec(uuid("{C0633BF6-2BDB-4FBD-A9D8-0BE19267D7D9}")) TDlgGuid),
                                         -1, -1, 56, 8,
                                         NULL,
                                         dialogItems,
                                         std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
        if(Dlg.get() == INVALID_HANDLE_VALUE)
        {
            return -1;
        }
        intptr_t askCode = startupInfo.DialogRun(Dlg.get());

        if(askCode != 4) return 0;

        FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(tmp_file_name) - 1, tmp_file_name };
        startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);
    }

    if(GetFileAttributesW(tmp_file_name)==0xFFFFFFFF)
        for(wchar_t *c=tmp_file_name; *c; c++)
        {
            if(*c!=' ')
            {
                for(; *c; c++)
                    if(*c==L'\\')
                    {
                        *c=0;
                        CreateDirectoryW(tmp_file_name, NULL);
                        *c=L'\\';
                    }
                CreateDirectoryW(tmp_file_name, NULL);
                break;
            }
        }
    if(*tmp_file_name && tmp_file_name[lstrlenW(tmp_file_name)-1] != L'\\') lstrcatW(tmp_file_name,L"\\");

    lstrcatW(tmp_file_name, PanelItem->FileName);

    //   
    WIN32_FIND_DATAW fdata;
    HANDLE h = FindFirstFileW(tmp_file_name, &fdata);
    if(h != INVALID_HANDLE_VALUE)
    {
        //     
        FindClose(h);

        const wchar_t *msgItems[8];
        msgItems[0] = L"Warning";//GetMsg(MWarning);
        msgItems[1] = L"File already exists";//GetMsg(MAlreadyExists);
        msgItems[2] = tmp_file_name;
        msgItems[3] = L"\x001";
        msgItems[5] = L"\x001";
        msgItems[6] = L"Overwrite";//GetMsg(MOverwrite);
        msgItems[7] = L"Cancel";//GetMsg(MCancel);
        wchar_t shortName[51];
        msgItems[2] = truncPathStr(shortName, tmp_file_name, 50);

        wchar_t param[50];
        if(fdata.ftLastWriteTime.dwLowDateTime == 0 &&
            fdata.ftLastWriteTime.dwHighDateTime == 0)
        {
            wsprintfW(param, L"Destination        %d"/*GetMsg(MDestination1)*/, fdata.nFileSizeLow);
        }
        else
        {
            FILETIME lastWriteTime;
            FileTimeToLocalFileTime(&fdata.ftLastWriteTime, &lastWriteTime);

            SYSTEMTIME time;
            FileTimeToSystemTime(&lastWriteTime, &time);

            wsprintfW(param, L"Destination        %d %02d.%02d.%04d %02d:%02d:%02d"/*GetMsg(MDestination2)*/,
                fdata.nFileSizeLow,
                time.wDay, time.wMonth, time.wYear,
                time.wHour, time.wMinute, time.wSecond);
        }
        msgItems[4] = param;
        int askCode = messageBox(FMSG_WARNING /*| FMSG_DOWN*/,
            msgItems,
            sizeof(msgItems)/sizeof(msgItems[0]), 2);
        if(askCode)
        {
            return 0;
        }
    }
    HANDLE out = CreateFileW(tmp_file_name, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

    if(out == INVALID_HANDLE_VALUE)
    {
        const wchar_t *msgItems[4];
        msgItems[0] = L"Error";//GetMsg(MError);
        msgItems[1] = L"Can not create file";//GetMsg(MCanNotCreate);
        msgItems[2] = tmp_file_name;
        msgItems[3] = L"Ok";//GetMsg(MOk);
        messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 1);
        return 0;
    }

    DWORD noBytesWritten;
    auto zx = (zxHandle*)Info->hPanel;
    WriteFile(out, zx->data, zx->dataSize, &noBytesWritten, 0);

    CloseHandle(out);

    return 1;
}

HANDLE WINAPI _export AnalyseW(const struct AnalyseInfo *Info)
{
    auto Name{ Info->FileName };
    auto DataSize{ Info->BufferSize };
    auto Data{ (const UCHAR *)Info->Buffer };
    if(DataSize < sizeof(HoHdr)+256)
        return nullptr;

    if(DataSize < (sizeof(HoHdr) + ((HoHdr*)Data)->noSecs*256) ||
        ((HoHdr*)Data)->checkSum != calculateCheckSum(*(HoHdr *)Data))
        return nullptr;

    DetectorResult result{};
    result.structSize = sizeof(result);

    xlook::Detector* dtr = manager->find(result, Name, Data, DataSize);

    if(result.action == do_extract && dtr && dtr->extract)
    {
        zxHandle *hPlugin = new zxHandle{};
        lstrcpyW(hPlugin->fileName, Name);

        hPlugin->dataSize = 65536;
        hPlugin->data = (BYTE*)malloc(hPlugin->dataSize);

        bool extractResult = dtr->extract((HoHdr*)Data, Data+sizeof(HoHdr), DataSize-sizeof(HoHdr), hPlugin->data, &hPlugin->dataSize, result.internal_id);
        if(!extractResult && hPlugin->dataSize > 65536)
        {
            hPlugin->data = (BYTE*)realloc(hPlugin->data, hPlugin->dataSize);
            extractResult = dtr->extract((HoHdr*)Data, Data+sizeof(HoHdr), DataSize-sizeof(HoHdr), hPlugin->data, &hPlugin->dataSize, result.internal_id);
        }

        if(!extractResult)
        {
            free(hPlugin->data);
            return nullptr;
        }

        lstrcpyW(hPlugin->ext, result.ext);
        std::copy_n(Data, 8, hPlugin->name);
        hPlugin->name[8] = 0;

        return (HANDLE)hPlugin;
    }

    return nullptr;
}

void WINAPI _export CloseAnalyseW(const struct CloseAnalyseInfo *Info)
{
    auto zx = (zxHandle*)Info->Handle;
    free(zx->data);
    delete zx;
}

HANDLE WINAPI _export OpenW(const struct OpenInfo *Info)
{
    if(Info->OpenFrom == OPEN_ANALYSE)
    {
        OpenAnalyseInfo *An = (OpenAnalyseInfo *)Info->Data;
        return An->Handle;
    }

    if(Info->OpenFrom != OPEN_COMMANDLINE) return nullptr;
    const OpenCommandLineInfo *CommandLineInfo = (const OpenCommandLineInfo *)Info->Data;
    const wchar_t *p = CommandLineInfo->CommandLine;

    wchar_t prefix  [10];
    wchar_t cmd_line[300];

    for(int i = 0; i < lstrlenW(p); ++i)
    {
        if(p[i] == ':')
        {
            const wchar_t *p2 = p + i + 1;
            while(*p2 && *p2 == ' ') ++p2; // skip spaces

            lstrcpyW(cmd_line, p2);
            prefix[i] = 0;
            break;
        }
        prefix[i] = p[i];
    }

    bool view = !lstrcmpW(prefix, open_in_viewer_prefix);
    bool edit = !lstrcmpW(prefix, open_in_editor_prefix);
    /*bool save = !lstrcmp(prefix, export_to_file_prefix);*/

    FSF.Unquote(cmd_line);

    HANDLE file = CreateFileW(cmd_line, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, 0, 0);

    // file not found
    if(file == INVALID_HANDLE_VALUE) return nullptr;

    DWORD read  = 0;
    DWORD write = 0;

    HoHdr hdr;

    DWORD file_size = GetFileSize(file, NULL);
    ReadFile(file, &hdr, sizeof(hdr), &read, 0);

    SetFilePointer(file, 0, 0, FILE_BEGIN);

    // it is not HoBeta file :(
    if(file_size < (sizeof(HoHdr) + 256) ||
        file_size < (sizeof(HoHdr) + hdr.noSecs*256) ||
        hdr.checkSum != calculateCheckSum(hdr))
    {
        CloseHandle(file);
        //errorMsg("not hobeta file");

        if(view)
            startupInfo.Viewer(cmd_line, nullptr, 0, 0, -1, -1, VF_NONE, 866);
        if(edit)
            startupInfo.Editor(cmd_line, nullptr, 0, 0, -1, -1, EN_NONE, -1, -1, 866);
        /*if(save)
        {
            // !!! show warning that export isn't supported for regular files
        }*/
        return nullptr;
    }

    BYTE* data = (BYTE*)malloc(file_size+10000);
    ReadFile(file, data, file_size, &read, 0);
    CloseHandle(file);

    DetectorResult result{};
    result.structSize = sizeof(result);

    xlook::Detector* dtr = manager->find(result, cmd_line, data, file_size);

    wchar_t title        [300];
    wchar_t tmp_file_name[300];

    makeTrDosName(title, data);

    /*if(save)
    {
        FarDialogItem item[6];
        item[0] = InitItem(DI_DOUBLEBOX, 3, 1, 52, 6, FALSE, 0, 0, 0, "xLook: export");
        item[1] = InitItem(DI_TEXT, 5, 2, 0, 0, 0, 0, 0, 0, "Export to:");
        item[2] = InitItem(DI_EDIT, 5, 3, 50, 0, TRUE, "xLook_export_to", DIF_HISTORY | DIF_EDITEXPAND, 0, "");
        item[3] = InitItem(DI_TEXT, 0, 4, 0, 0, 0, 0, DIF_SEPARATOR, 0, 0);

        item[4] = InitItem(DI_BUTTON, 11, 5, 0, 0, 0, 0, DIF_CENTERGROUP, TRUE, "&Ok");
        item[5] = InitItem(DI_BUTTON, 35, 5, 0, 0, 0, 0, DIF_CENTERGROUP, 0, "Ca&ncel");

        if(startup_info.Dialog(startup_info.ModuleNumber, -1, -1, 56, 8, 0, item, 6) != 4)
        {
            free(data);
            return INVALID_HANDLE_VALUE;
        }
        lstrcpy(tmp_file_name, item[2].Data);
    }
    else*/
    {
        tmp_file_name[0] = 0;
        GetTempFileNameW(tmp_folder, L"!xv", 0, tmp_file_name);
        DeleteFileW(tmp_file_name);

        //read_associations(startup_info.ModuleName);

        lstrcatW(tmp_file_name, L"\\");
        CreateDirectoryW(tmp_file_name, 0);

        makePCName((const char*)data, tmp_file_name + lstrlenW(tmp_file_name));

        if(dtr && result.ext[0])
            lstrcatW(tmp_file_name, result.ext);
        else
            lstrcatW(tmp_file_name, L".bin");

/*      GetTempFileName(tmp_folder, "!xv", 0, tmp_file_name);

        if(dtr && result.ext[0])
        {
            char old_name[300];
            lstrcpy(old_name, tmp_file_name);
            lstrcat(tmp_file_name, result.ext);
            MoveFile(old_name, tmp_file_name);
        }*/
    }

    HANDLE out = CreateFileW(tmp_file_name, GENERIC_WRITE, FILE_SHARE_READ, 0, CREATE_ALWAYS, 0, 0);

    // extract
    bool isExtracted = false, extractResult = true;
    if(result.action == do_extract && dtr && dtr->extract)
        extractResult = isExtracted = extract(dtr, data, file_size, out, result.internal_id);
    if(!isExtracted)
    {
        // strip HoBeta header
        if((dtr && result.action == do_strip_hobeta) || strip_hobeta_header/* || save*/)
        {
            // strip garbage after file body if file length in the HoBeta header is ok.
            if(hdr.size >= 256*(hdr.noSecs - 1) && hdr.size <= 256*(hdr.noSecs))
                WriteFile(out, data + sizeof(HoHdr), hdr.size, &write, 0);
            else
                WriteFile(out, data + sizeof(HoHdr), file_size - sizeof(HoHdr), &write, 0);
        }
        else
            // leave original file
            WriteFile(out, data, file_size, &write, 0);
    }
    CloseHandle(out);
    free(data);

    // update title
    if(dtr && result.title[0] && extractResult)
        lstrcatW(title, result.title);
    else
        lstrcatW(title, L"[HoBeta]");

    if(dtr && result.format[0] && extractResult)
    {
        for(int i = 0; i < _no_assosiations; ++i)
        {
            Association& cur = _associations[i];
            if(!lstrcmpW(result.format, cur.format))
            {
                // association found!
                wchar_t  external_viewer[300];
                wchar_t* to = external_viewer;

                wchar_t* command_line_template = 0;

                if(view) command_line_template = cur.view;
                if(edit) command_line_template = cur.edit;
                
                if(command_line_template && *command_line_template)
                {   
                    // command line defined
                    wchar_t* p = command_line_template;
                    // replace %f with real file name
                    while(*p)
                    {
                        if(p[0] == L'%' && p[1] == L'f')
                        {
                            lstrcpyW(to, tmp_file_name);
                            to += lstrlenW(tmp_file_name);
                            p += 2;
                        }
                        else
                            *to++ = *p++;
                    }

                    STARTUPINFOW info{};
                    PROCESS_INFORMATION proc_info;

                    if(CreateProcessW(0, external_viewer, 0, 0, TRUE, 0, 0, 0, &info, &proc_info))
                        return nullptr;

                    // can't execute external viewer. so fall in internal viewer/editor
                    // !!! show warning
                }
            }
        }
    }

    // use internal viewer/editor
    DWORD flags = EF_DISABLEHISTORY | EF_NONMODAL | EF_ENABLE_F6 | EF_DELETEONLYFILEONCLOSE;

    if(view)
        startupInfo.Viewer(tmp_file_name, title, 0, 0, -1, -1, flags, 866);
    if(edit)
        startupInfo.Editor(tmp_file_name, title, 0, 0, -1, -1, flags, -1, -1, 866);

    return nullptr;
}

void WINAPI _export GetGlobalInfoW(struct GlobalInfo *Info)
{
    Info->StructSize = sizeof(struct GlobalInfo);
    Info->MinFarVersion = MAKEFARVERSION(3, 0, 0, 2927, VS_RELEASE);   //   Far Manager,    API
    Info->Version = MAKEFARVERSION(0, 2, 0, 1, VS_RC);   //   0.2.0.1, -
    Info->Guid = MainGuid;
    Info->Title = L"xLOOK";
    Info->Description = L"xLOOK";
    Info->Author = L"Elf/2 & Alexandr Medvedev";
}
