#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <iterator>
#include "plugin.hpp"

#include "types.hpp"
#include "types_trd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Options           op;

namespace
{
  enum Action { ASK_USER, DELETE_ALL };
  Action userAction;

  bool   keepSilence;
}

void Manager::move(void)
{
  if(noDelFiles == 0 && noDelFolders == 0) return;

  int  fromSec, fromTrk, toSec, toTrk;

  bool delFileFound = false;
  int  noDelSecs = 0;
  int  toFileNum = 0;
  
  if(noDelFiles != 0)
  {
    for(int i = 0; i < noFiles; ++i)
    {
      if(files[i].name[0] == 0x01)
      {
        noDelSecs += files[i].noSecs;
        if(!delFileFound)
        {
          delFileFound = true;
          toTrk = files[i].trk;
          toSec = files[i].sec;
        }
      }
      else
      {
        if(delFileFound)
        {
          fromTrk = files[i].trk;
          fromSec = files[i].sec;
          files[i].trk = toTrk;
          files[i].sec = toSec;
          int noSecs = files[i].noSecs;
          //  
          for(int j = 0; j < noSecs; ++j)
          {
            BYTE buf[sectorSize];
            read (fromTrk, fromSec, buf);
            write(toTrk,   toSec,   buf);
            
            if(++fromSec == 16) { fromTrk++; fromSec = 0; }
            if(++toSec   == 16) { toTrk++;   toSec = 0; }
          }
          memcpy(&files[toFileNum], &files[i], sizeof(FileHdr));
          memcpy(&pcFiles[toFileNum], &pcFiles[i], sizeof(ExtFileHdr));
          if(dsOk) fileMap[toFileNum] = fileMap[i];
        }
        ++toFileNum;
      }
    }

    if(dsOk) for(int i = 0; i < noDelFiles; ++i) fileMap[noFiles-noDelFiles+i] = 0;
    noFiles     -= noDelFiles;
    noDelFiles   = 0;
    diskInfo.firstFreeTrk = toTrk;
    diskInfo.firstFreeSec = toSec;
    diskInfo.noFiles      = noFiles;
    diskInfo.noDelFiles   = noDelFiles;
    diskInfo.noFreeSecs  += noDelSecs;
  }
  if(dsOk && noDelFolders != 0)
  {
    int no_folders = noFolders;
    for(int i = 0; i < no_folders; ++i)
    {
      if(folders[i][0] == 0x01)
      {
        if(curFolderNum == i+1)
        {
          curFolderNum = 0; //    
          *curFolder   = 0; //    
        }
        if(curFolderNum > i+1) --curFolderNum;
        for(int j = 0; j < noFiles; ++j)
          if(fileMap[j] >= i+1) --fileMap[j];
        for(int j = 0; j < noFolders; ++j)
          if(folderMap[j] >= i+1) --folderMap[j];

        memcpy(&folders[i], &folders[i+1], 11*(noFolders-i-1));
        memcpy(folderMap+i, folderMap+i+1, noFolders-i-1);
        --i;
        --no_folders;
      }
    }
    for(int i = 0; i < noDelFolders; ++i) folderMap[noFolders-noDelFolders+i] = 0;
    noFolders   -= noDelFolders;
    noDelFolders = 0;
  }
}

ExitCode Manager::markFolder(int fNum)
{
  if(userAction == ASK_USER && !keepSilence)
  {
    int noItems = 0; //      
    for(int i = 0; i < noFiles; ++i)
      if(fileMap[i] == fNum) ++noItems;
    for(int i = 0; i < noFolders; ++i)
      if(folderMap[i] == fNum) ++noItems;
    
    if(noItems > 0) //    ,  ...
    {
      const wchar_t *msgItems[]
      {
          GetMsg(MDeleteFolder), GetMsg(MDelFolder),
          nullptr,
          GetMsg(MDeleteButton), GetMsg(MAll), GetMsg(MSkip), GetMsg(MCancel)
      };
      
      wchar_t folderName[300];
      wchar_t temp[300] = L"";
      int  parentFolder = fNum;
      while(parentFolder)
      {
        if(temp[0])
          wsprintfW(folderName, L"%s\\%s", pcFolders[parentFolder-1], temp);
        else
          wsprintfW(folderName, L"%s", pcFolders[parentFolder-1]);
        lstrcpyW(temp, folderName);
        parentFolder = folderMap[parentFolder-1];
      }
      
      wchar_t shortName[50];
      msgItems[2] = truncPathStr(shortName, folderName, 40);
      int button = messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 4);
      switch(button)
      {
        case -1:
        case 3:
                  return CANCEL;
        case 2:
                  return SKIP;
        case 1:
                  userAction = DELETE_ALL;
      }
    }
  }
  
  for(int i = 0; i < noFiles; ++i)
  {
    if(fileMap[i] == fNum)
    {
      if(files[i].name[0] != 0x01) ++noDelFiles;
      files[i].name[0] = 0x01;
    }
  }
  ExitCode exitCode = OK;
  for(int i = 0; i < noFolders; ++i)
    if(folderMap[i] == fNum)
    {
      ExitCode code = markFolder(i+1);
      if(code != OK) exitCode = code;
      if(code == CANCEL) break;
    }
  if(exitCode == OK)
  {
    if(folders[fNum-1][0] != 0x01) ++noDelFolders;
    folders[fNum-1][0] = 0x01;
  }
  return exitCode;
}

int Manager::deleteFiles(PluginPanelItem *panelItem, int noItems, int opMode)
{
  if(noItems == 0) return TRUE;
  if(fmt->isProtected(img))
  {
    const wchar_t * const msgItems[]{ GetMsg(MWarning), GetMsg(MCanNotDel), GetMsg(MWriteProtected), GetMsg(MOk) };
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    return FALSE;
  }

  //     
  //   
  readInfo();

  keepSilence = opMode & OPM_SILENT;
  userAction  = ASK_USER;
  
  if(!keepSilence)
  {
    const wchar_t *msgItems[5] { GetMsg(MDelete), GetMsg(MDeleteAsk), nullptr, GetMsg(MDeleteButton), GetMsg(MCancel) };

    wchar_t msg[30];
    if(noItems == 1)
      wsprintfW(msg, L"%s", panelItem[0].FileName);
    else
      wsprintfW(msg, GetMsg(MDelFiles), noItems);

    msgItems[2] = msg;
    if(messageBox(0, msgItems, std::size(msgItems), 2)!=0)
      return FALSE;
    if(noItems > 1)
    {
      const wchar_t *msgItems[5]{ GetMsg(MDeleteFiles), GetMsg(MDeleteAsk), nullptr, GetMsg(MAll), GetMsg(MCancel) };

      wchar_t msg[30];
      wsprintfW(msg, GetMsg(MDelFiles), noItems);
      msgItems[2] = msg;
      if(messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 2)!=0)
        return FALSE;
    }
  }
  if(!openHostFile()) return FALSE;

  bool allOk = true;
  
  for(int i = 0; i < noItems; ++i)
  {
    if(panelItem[i].FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    {
      int j = 0;
      for(; j < noFolders; ++j)
      {
        if(folderMap[j] != curFolderNum) continue;
        if(!lstrcmpW(pcFolders[j], panelItem[i].FileName)) break;
      }
      ExitCode code = markFolder(j+1);
      if(code != OK) allOk = false;
      if(code == CANCEL) break;
    }
    else
    {
      for(int j = 0; j < noFiles; ++j)
      {
        if(!lstrcmpW(pcFiles[j].name, panelItem[i].FileName))
        {
          if(files[j].name[0] != 0x01) ++noDelFiles;
          files[j].name[0] = 0x01;
          break;
        }
      }
    }
  }
  if(op.autoMove)
    move();
  else
    diskInfo.noDelFiles = noDelFiles;

  writeInfo();
  closeHostFile();
  op.reread = true;
  startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
  op.reread = false;
  return allOk;
}
