#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <wchar.h>
#include <memory>
#include <string>
#include "plugin.hpp"

#include "types.hpp"
#include "types_trd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Options           op;
extern const GUID MainGuid;

int Manager::setDirectory(const wchar_t* dirName, OPERATION_MODES opMode)
{
  if(!op.useDS || !dsOk) return FALSE;

  if(!lstrcmpW(dirName, L"\\"))
  {
    *curFolder   = 0;
    curFolderNum = 0;
    return TRUE;
  }
  if(!lstrcmpW(dirName, L".."))
  {
    if(*curFolder == 0) return FALSE;

    wchar_t* slash = wcsrchr(curFolder, L'\\');
    if(slash)
    {
      *slash       = 0;
      curFolderNum = folderMap[curFolderNum-1];
    }
    else
    {
      *curFolder   = 0;
      curFolderNum = 0;
    }
  }
  else
  {
    int  fNum = 0;
    for(;fNum < noFolders; ++fNum)
    {
      if(folderMap[fNum] != curFolderNum) continue;
      if(!lstrcmpW(pcFolders[fNum], dirName)) break;
    }
    if(fNum == noFolders) return FALSE;

    curFolderNum = fNum + 1;
    lstrcatW(curFolder, L"\\");
    lstrcatW(curFolder, dirName);
  }
  return TRUE;
}

int Manager::makeDirectory(const wchar_t* dirName, OPERATION_MODES opMode)
{
  if(!op.useDS) return -1;
  if(fmt->isProtected(img))
  {
    const wchar_t * const msgItems[]{ GetMsg(MWarning), GetMsg(MCanNotMakeDir), GetMsg(MWriteProtected), GetMsg(MOk) };
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    return FALSE;
  }
  if(!dsOk)
  {
    const wchar_t * const msgItems[]{ GetMsg(MWarning), GetMsg(MDSNotInstalled), GetMsg(MInstallDS), GetMsg(MOk),GetMsg(MCancel) };
    if(messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 2) != 0) return -1;
    dsOk = true;
  }
  if(noFolders == 127)
  {
    errorMessage(GetMsg(MTooManyFolders));
    return -1;
  }
  
  const FarDialogItem dialogItems[] =
  {
    { DI_DOUBLEBOX, 3, 1, 60, 6, { FALSE },             nullptr, nullptr, 0,                                   GetMsg(MMakeFolder),   0, 0, { 0, 0 } },
    { DI_TEXT,      5, 2,  0, 0, { FALSE },             nullptr, nullptr, 0,                                   GetMsg(MCreateFolder), 0, 0, { 0, 0 } },
    { DI_EDIT,      5, 3, 58, 0, { TRUE },  L"xTRD_folder_name", nullptr, DIF_HISTORY,                         dirName,               0, 0, { 0, 0 } },
    { DI_TEXT,      3, 4,  0, 0, { FALSE },             nullptr, nullptr, DIF_BOXCOLOR | DIF_SEPARATOR,        L"",                  12, 0, { 0, 0 } },
    { DI_BUTTON,    0, 5,  0, 0, { FALSE },             nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MOk),           0, 0, { 0, 0 } },
    { DI_BUTTON,    0, 5,  0, 0, { FALSE },             nullptr, nullptr, DIF_CENTERGROUP,                     GetMsg(MCancel),       0, 0, { 0, 0 } }
  };
  
  //    ,  
  std::wstring DirNameStorage;
  if((opMode & OPM_SILENT) == 0)
  {
      std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                     &__uuidof(struct __declspec(uuid("{B8D86961-59D4-4A83-8C6F-99FAD3E32FF0}")) TDlgGuid),
                                     -1, -1, 64, 8,
                                     nullptr,
                                     dialogItems,
                                     std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
    if(Dlg.get() == INVALID_HANDLE_VALUE)
    {
        return FALSE;
    }
    intptr_t askCode = startupInfo.DialogRun(Dlg.get());

    if(askCode != 4) return -1;

    auto Sz = startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, nullptr);
    DirNameStorage.resize(Sz);

    FarDialogItemData ItemData = { sizeof(FarDialogItemData), DirNameStorage.size(), DirNameStorage.data() };
    startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);
    dirName = DirNameStorage.c_str();
  }
  if(*dirName == 0) return 1;

  if(!openHostFile()) return 0;
  std::fill(std::begin(folders[noFolders]), std::end(folders[noFolders]), ' ');

  make8x3name(dirName, folders[noFolders]);
  if(folders[noFolders][0] == 0x01) ++noDelFolders;
  folderMap[noFolders] = curFolderNum;
  ++noFolders;

  writeInfo();
  closeHostFile();
  startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
  return 1;
}
