#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <wchar.h>
#include <memory>
#include <algorithm>
#include "plugin.hpp"

#include "types.hpp"
#include "types_trd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Options           op;
extern const GUID MainGuid;

const DWORD PKF_ANY = (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED) | SHIFT_PRESSED | (LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED);

static wchar_t* appendFileName(const wchar_t* fileName, wchar_t* to, bool shortName)
{
  wchar_t buf[300];
  if(shortName)
    GetShortPathNameW(fileName, buf, sizeof(buf));
  else
  {
    lstrcpyW(buf, fileName);
    quoteSpaceOnly(buf);
  }
  
  wmemcpy(to, buf, lstrlenW(buf));
  return to+lstrlenW(buf);
}

int Manager::processKey(const INPUT_RECORD *Rec)
{
    if(Rec->EventType != KEY_EVENT)
    {
        return FALSE;
    }

    if(!Rec->Event.KeyEvent.bKeyDown)
    {
        return FALSE;
    }

  auto key = Rec->Event.KeyEvent.wVirtualKeyCode;
  auto controlState = Rec->Event.KeyEvent.dwControlKeyState;

  if(key == VK_RETURN)
  {
    wchar_t buf[1024];
    startupInfo.PanelControl(this,FCTL_GETCMDLINE, 0, buf);
    if(*buf) return FALSE;
    
    HANDLE hScreen = startupInfo.SaveScreen(0,0,-1,-1);

    intptr_t Size = startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, nullptr);
    std::unique_ptr<PluginPanelItem, decltype(::free) *> Item((PluginPanelItem *)::malloc(Size), ::free);
    FarGetPluginPanelItem GetItem = { sizeof(FarGetPluginPanelItem), (size_t)Size, Item.get() };
    intptr_t r = startupInfo.PanelControl(this,FCTL_GETCURRENTPANELITEM, 0, &GetItem);
    startupInfo.RestoreScreen(hScreen);

    if(r == FALSE)
    {
        return FALSE;
    }
    const wchar_t* name = Item->FileName;
    int fNum = 0;
    for(; fNum < noFiles; ++fNum)
      if(!lstrcmpW(name, pcFiles[fNum].name)) break;

    if(fNum == noFiles) return FALSE;

    wchar_t  cmdLine[600];
    wchar_t* p          = op.cmdLine1;
    wchar_t* type       = op.types1;
    bool  fullScreen = op.fullScreen1;

    if(controlState & (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED))
    {
      p          = op.cmdLine2;
      type       = op.types2;
      fullScreen = op.fullScreen2;
    }

    wchar_t* ptr      = pcFiles[fNum].name;
    wchar_t  fileType = L' ';
    
    if(*type == '*') goto OkType;
    
    while(*ptr && *ptr != '.') ++ptr;
    if(*ptr) fileType = *(ptr+2);
      
    while(*type)
    {
      if(*type == fileType) goto OkType;
      ++type;
    }
    return FALSE;
OkType:
    wchar_t* to = cmdLine;
    while(*p)
    {
      if(*p != L'%')
        *to++ = *p++;
      else
      {
        switch(*(p+1))
        {
          case L'P':
                    to = appendFileName(hostFileName, to, false);
                    break;
          case L'p':
                    to = appendFileName(hostFileName, to, true);
                    break;
          case L'D':
                    to = appendFileName(pointToName(hostFileName), to, false);
                    break;
          case L'd':
                    to = appendFileName(pointToName(hostFileName), to, true);
                    break;
          case L'f':
                    std::transform(std::cbegin(files[fNum].name), std::cend(files[fNum].name), to, Cp866ToUcs2);
                    to += std::size(files[fNum].name);
                    *to++ = L'.';
                    *to++ = files[fNum].type;
                    break;
          case L'N':
                    to += wsprintfW(to, L"%d", fNum+1);
                    break;
          case L'n':
                    to += wsprintfW(to, L"%d", fNum);
                    break;
          default:
                    *to++ = *p++;
                     p   -= 2;
        }
        p += 2;
      }
      if(to - cmdLine >= 300) break;
    }
    *to = 0;
    
    STARTUPINFOW sInfo{};
    if(fullScreen) sInfo.dwFlags = STARTF_RUNFULLSCREEN;
    
    PROCESS_INFORMATION procInfo;
    if(!CreateProcessW(NULL, cmdLine, NULL, NULL, FALSE, CREATE_NEW_CONSOLE, NULL, NULL, &sInfo, &procInfo))
    {
      const wchar_t * const msgItems[]{ GetMsg(MxPluginWarning), GetMsg(MCanNotExec), cmdLine, GetMsg(MOk) };
      messageBox(FMSG_WARNING /*| FMSG_DOWN*/, msgItems, std::size(msgItems), 1);
    }

    return TRUE;
  }
  if(((controlState & SHIFT_PRESSED) && key == VK_F6) ||
     (!(controlState & PKF_ANY) && key == VK_F4) ||
     ((controlState & (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED)) && key == 0x41))
  {
    if(fmt->isProtected(img))
    {
      const wchar_t * const msgItems[]{ GetMsg(MWarning), GetMsg(MCanNotEdit), GetMsg(MWriteProtected), GetMsg(MOk) };
      messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
      return TRUE;
    }

    HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);
    PanelInfo info = { sizeof(PanelInfo) };
    startupInfo.PanelControl(this, FCTL_GETPANELINFO, 0, &info);
    startupInfo.RestoreScreen(screen);
    
    if(info.ItemsNumber > 0 && info.CurrentItem > 0)
    {
      int fNum = 0;
      intptr_t Size = startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, nullptr);
      std::unique_ptr<PluginPanelItem, decltype(::free)*> Item((PluginPanelItem *)::malloc(Size), ::free);
      FarGetPluginPanelItem GetItem = { sizeof(FarGetPluginPanelItem), (size_t)Size, Item.get() };
      startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, &GetItem);

      const wchar_t *fName = Item->FileName;
      if(Item->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      { //  
        for(fNum = 0; fNum < noFolders; ++fNum)
        {
          if(folderMap[fNum] != curFolderNum) continue;
          if(!lstrcmpW(fName, pcFolders[fNum])) break;
        }

        wchar_t name[12] = L"           ";

        std::transform(std::cbegin(folders[fNum]), std::cend(folders[fNum]), name, Cp866ToUcs2);

        int isDeleted = (name[0] == 0x01) ? 1 : 0;

        const FarDialogItem dialogItems[]=
        {
          { DI_DOUBLEBOX, 3, 1, 34, 6, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MRenameFolder), 0, 0, { 0, 0 } },
          { DI_TEXT,      5, 2,  0, 0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MFolderName),   0, 0, { 0, 0 } },
          { DI_EDIT,     13, 3, 24, 0, { TRUE },  nullptr, nullptr,                                   0, name,                  std::size(name) - 1, 0, { 0, 0 } },
          { DI_TEXT,      3, 4,  0, 0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"",                   0, 0, { 0, 0 } },
          { DI_BUTTON,    0, 5,  0, 0, { FALSE }, nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MOk),           0, 0, { 0, 0 } },
          { DI_BUTTON,    0, 5,  0, 0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel),       0, 0, { 0, 0 } }
        };

        std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                         &__uuidof(struct __declspec(uuid("{3FA96C24-F5E1-427C-9589-DCF70458091E}")) TDlgGuid),
                                         -1, -1, 38, 8,
                                         NULL,
                                         dialogItems,
                                         std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
        if(Dlg.get() == INVALID_HANDLE_VALUE)
        {
            return FALSE;
        }
        intptr_t askCode = startupInfo.DialogRun(Dlg.get());

        if(askCode != 4) return TRUE;

        FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(name) - 1, name };
        startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);

        std::fill(std::begin(folders[fNum]), std::end(folders[fNum]), ' ');

        {
            unsigned i = 0;
            for(auto c : name)
            {
                if(c == 0) break;
                //   cp866
                WideCharToMultiByte(866, 0, &c, 1, &folders[fNum][i++], 1, nullptr, nullptr);
            }
        }

        int isDeletedAfter = (folders[fNum][0] == 0x01) ? 1 : 0;
        switch(isDeleted - isDeletedAfter)
        {
          case 1:
                    //  
                    noDelFolders--;
                    break;
          case -1:
                    folders[fNum][0] = '?';
                    break;
        }
      }
      else
      { //   
          for(; fNum < noFiles; ++fNum)
          {
              if(!lstrcmpW(fName, pcFiles[fNum].name))
              {
                  break;
              }
          }

        wchar_t name[9] = L"        ";
        std::transform(std::cbegin(files[fNum].name), std::cend(files[fNum].name), name, Cp866ToUcs2);

        int isDeleted = (name[0] == 0x01) ? 1 : 0;
        wchar_t type[2] = L" ";
        type[0] = files[fNum].type ? files[fNum].type : ' ';

        wchar_t start[6];
        wsprintfW(start, L"%d", files[fNum].start);

        const FarDialogItem dialogItems[]=
        {
          { DI_DOUBLEBOX,  3, 1, 34, 7, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MEditFileInfo), 0, 0, { 0, 0 } },
          { DI_TEXT,       5, 2,  0, 0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MFileName),     0, 0, { 0, 0 } },
          { DI_EDIT,      24, 2, 32, 0, { TRUE },  nullptr, nullptr,                                   0, name,                  std::size(name) - 1, 0, { 0, 0 } },
          { DI_TEXT,       5, 3,  0, 0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MFileType),     0, 0, { 0, 0 } },
          { DI_EDIT,      24, 3, 25, 0, { FALSE }, nullptr, nullptr,                                   0, type,                  std::size(type) - 1, 0, { 0, 0 } },
          { DI_TEXT,       5, 4,  0, 0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MStartAddress), 0, 0, { 0, 0 } },
          { DI_EDIT,      24, 4, 29, 0, { FALSE }, nullptr, nullptr,                                   0, start,                 std::size(start) - 1, 0, { 0, 0 } },
          { DI_TEXT,       3, 5,  0, 0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"",                   0, 0, { 0, 0 } },
          { DI_BUTTON,     0, 6,  0, 0, { FALSE }, nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MOk),           0, 0, { 0, 0 } },
          { DI_BUTTON,     0, 6,  0, 0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel),       0, 0, { 0, 0 } }
        };

        std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                         &__uuidof(struct __declspec(uuid("{B9F26504-0081-441E-B9EA-5841B11DE104}")) TDlgGuid),
                                         -1, -1, 38, 9,
                                         nullptr,
                                         dialogItems,
                                         std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
        if(Dlg.get() == INVALID_HANDLE_VALUE)
        {
            return FALSE;
        }
        intptr_t askCode = startupInfo.DialogRun(Dlg.get());

        if(askCode != 8) return TRUE;

        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(name) - 1, name };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);
        }

        std::fill(std::begin(files[fNum].name), std::end(files[fNum].name), ' ');

        {
            unsigned i = 0;
            for(auto c : name)
            {
                if(c == 0) break;
                //   cp866
                WideCharToMultiByte(866, 0, &c, 1, &files[fNum].name[i++], 1, nullptr, nullptr);
            }
        }

        int isDeletedAfter = (files[fNum].name[0] == 0x01) ? 1 : 0;
        switch(isDeleted - isDeletedAfter)
        {
          case 1:
                    //  
                    noDelFiles--;
                    diskInfo.noDelFiles = noDelFiles;
                    break;
          case -1:
                    files[fNum].name[0] = '?';
                    break;
        }

        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(type) - 1, type };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 4, &ItemData);
        }

        files[fNum].type = 'C';
        if(type[0] > 0) files[fNum].type = type[0];

        {
            FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(start) - 1, start };
            startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 6, &ItemData);
        }

        int result = str2int(start, std::size(start) - 1);
        if(result != -1) files[fNum].start = result;
      }

      if(!openHostFile()) return TRUE;
      writeInfo();
      closeHostFile();

      startupInfo.PanelControl(this, FCTL_UPDATEPANEL, 1, nullptr);
      startupInfo.PanelControl(this, FCTL_REDRAWPANEL, 0, nullptr);
      startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
      return TRUE;
    }
  }
  if(key == VK_F2 && !(controlState & PKF_ANY))
  {
    diskMenu();
    return TRUE;
  }
  return FALSE;
}

bool Manager::checkDisk(void)
{
  if(diskInfo.noFiles    != noFiles) return true;
  if(diskInfo.noDelFiles != noDelFiles) return true;

  int totalSecs = 0;
  switch(diskInfo.type)
  {
    case 0x16:
              totalSecs = 159*16;
              break;
    case 0x17:
    case 0x18:
              totalSecs = 79*16;
              break;
    case 0x19:
              totalSecs = 39*16;
              break;
  }

  int trk = 1;
  int sec = 0;
  int noTotalSecs = 0;
  for(int i = 0; i < noFiles; ++i)
  {
    int noSecs = files[i].noSecs;
    noTotalSecs += noSecs;
    
    if(files[i].trk != trk || files[i].sec != sec) return true;
    trk += noSecs / 16;
    sec += noSecs % 16;
    if(sec > 15) { ++trk; sec %= 16; }
  }
  if(diskInfo.firstFreeTrk != trk || diskInfo.firstFreeSec != sec) return true;
  if(noTotalSecs + diskInfo.noFreeSecs != totalSecs) return true;
  
  return false;
}

bool Manager::checkEntry(char num, bool del, bool *checked)
{
  int i;

  for(i = 0; i < noFolders; i++)
    if(folderMap[i]==num)//    
    {
       bool isDel = folders[i][0] == 0x01;
       if(del && !isDel) return true;//    
       checked[i] = true;
       if(checkEntry(i + 1, isDel, checked)) return true;//  
    }

  return false;
}

bool Manager::checkDS(void)
{
  bool checked[127];
  int i;

  if(noFolders==0) return false; // 

  for(i=0; i < noFiles; i++)
    if(fileMap[i] > noFolders) return true;//   

//  for(i = 0; i < noFolders; i++)
//    if(folderMap[i] > noFolders) return true;//   

  for(i = 0; i < noFiles; i++)
    if((files[i].name[0] != 0x01) &&
       (folders[fileMap[i] - 1][0] == 0x01)) return true;//   

  for(i = 0; i < 127; i++)
    checked[i] = false;//  

  if(checkEntry(0, false, checked)) return true;// 

  for(i = 0; i < noFolders; i++)
    if(!checked[i]) return true;//   

  return false;
}

int Manager::diskMenu()
{
  FarMenuItem menuItems[5];
  ZeroMemory(menuItems, sizeof(menuItems));
  menuItems[0].Flags = MIF_SELECTED;

  int noMenuItems = 2;

  menuItems[0].Text = GetMsg(MCheckDisk);
  if(fmt->isProtected(img))
  {
    menuItems[1].Text = GetMsg(MRemoveProtection);
  }
  else
  {
    menuItems[1].Text = GetMsg(MProtectDisk);
    menuItems[2].Text = GetMsg(MEditDiskTitle);
    menuItems[3].Text = GetMsg(MMoveDisk);
    menuItems[4].Text = GetMsg(MCleanFreeSpace);
    noMenuItems = 5;
  }

  int exitCode = startupInfo.Menu(&MainGuid,
                                  &__uuidof(struct __declspec(uuid("{7E26FEFD-F625-4E6B-99A4-C00CA28E052A}")) TMenuGuid),
                                  -1,-1,
                                  0,
                                  FMENU_WRAPMODE,
                                  GetMsg(MDiskMenu),
                                  NULL, NULL, NULL, NULL,
                                  menuItems,
                                  noMenuItems);
  if(exitCode == -1) return FALSE;

  switch(exitCode)
  {
    case 0: /* check disk */
              {
                const wchar_t *msgItems[5];
                msgItems[0] = GetMsg(MCheckDisk2);
                msgItems[3] = L"\001";
                msgItems[4] = GetMsg(MOk);
                wchar_t msg1[30];
                if(checkDisk())
                  wsprintfW(msg1, GetMsg(MTestFiles), GetMsg(MErrors));
                else
                  wsprintfW(msg1, GetMsg(MTestFiles), GetMsg(MNoErrors));
                msgItems[1] = msg1;
                wchar_t msg2[30];
                if(!op.useDS)
                  wsprintfW(msg2, GetMsg(MTestDS), GetMsg(MDisabled));
                else
                  if(!dsOk)
                    wsprintfW(msg2, GetMsg(MTestDS), GetMsg(MAbsent));
                  else
                    if(checkDS())
                      wsprintfW(msg2, GetMsg(MTestDS), GetMsg(MErrors));
                    else
                      wsprintfW(msg2, GetMsg(MTestDS), GetMsg(MNoErrors));
                msgItems[2] = msg2;
                messageBox(0, msgItems, std::size(msgItems), 1);
                break;
              }
    case 1: /* set/remove protection */
              {
                bool ok;
                if(fmt->isProtected(img))
                  ok = fmt->protect(img, false);
                else
                  ok = fmt->protect(img, true);
                if(!ok)
                {
                  const wchar_t * const msgItems[]{ GetMsg(MError), GetMsg(MCanNotProtect), GetMsg(MOk) };
                  messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
                }
                else
                {
                  startupInfo.PanelControl(this,FCTL_UPDATEPANEL,1,nullptr);
                  startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
                }
                break;
              }
    case 2: /* edit disk title */
              {
                wchar_t title[12];
                std::transform(std::cbegin(diskInfo.title), std::cend(diskInfo.title), title, Cp866ToUcs2);
                title[11] = 0;
                
                const FarDialogItem dialogItems[] =
                {
                  { DI_DOUBLEBOX,  3, 1, 34, 6, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MEditTitle),   0, 0, { 0, 0 } },
                  { DI_TEXT,       5, 2,  0, 0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MChangeTitle), 0, 0, { 0, 0 } },
                  { DI_EDIT,      13, 3, 24, 0, { TRUE },  nullptr, nullptr,                                   0, title,                std::size(title) - 1, 0, { 0, 0 } },
                  { DI_TEXT,       3, 4,  0, 0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"",                  0, 0, { 0, 0 } },
                  { DI_BUTTON,     0, 5,  0, 0, { FALSE }, nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MOk),          0, 0, { 0, 0 } },
                  { DI_BUTTON,     0, 5,  0, 0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel),      0, 0, { 0, 0 } }
                };

                std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                                 &__uuidof(struct __declspec(uuid("{8BE60621-7C3F-46E9-ABD4-A5A5B202A925}")) TDlgGuid),
                                                 -1, -1, 38, 8,
                                                 NULL,
                                                 dialogItems,
                                                 std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
                if(Dlg.get() == INVALID_HANDLE_VALUE)
                {
                    return FALSE;
                }
                intptr_t askCode = startupInfo.DialogRun(Dlg.get());

                if(askCode != 4) return TRUE;

                std::fill(std::begin(diskInfo.title), std::end(diskInfo.title), ' ');

                FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(title) - 1, title };
                startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);

                {
                    unsigned i = 0;
                    for(auto c : title)
                    {
                        if(c == 0) break;
                        //   cp866
                        WideCharToMultiByte(866, 0, &c, 1, &diskInfo.title[i++], 1, nullptr, nullptr);
                    }
                }

                if(!openHostFile()) return FALSE;
                writeInfo();
                closeHostFile();

                startupInfo.PanelControl(this,FCTL_UPDATEPANEL,1,nullptr);
                startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
                break;
              }
    case 3: /* move disk */
              {
                if(noDelFiles == 0 && noDelFolders == 0) return TRUE;
                if(!openHostFile()) return FALSE;
                move();
                writeInfo();
                closeHostFile();

                startupInfo.PanelControl(this,FCTL_UPDATEPANEL,1, nullptr);
                startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);
                break;
              }
    case 4: /* clean free space */
              {
                BYTE zeroSec[sectorSize]{};
                if(!openHostFile()) return FALSE;
                
                int trk = diskInfo.firstFreeTrk;
                int sec = diskInfo.firstFreeSec;
                for(int i = 0; i < diskInfo.noFreeSecs; ++i)
                {
                  write(trk, sec, zeroSec);
                  if(++sec == 16) { sec = 0; ++trk; }
                  if(trk == 256) break;
                }
                closeHostFile();
                break;
              }
  }
  return TRUE;
}
